@@ -1,6 +1,6 @@
-# Authors: Murat Cihan Sorkun <m.c.sorkun@differ.nl>, Dajt Mullaj <d.mullaj@differ.nl>
+# Authors: Murat Cihan Sorkun <mcsorkun@gmail.com>, Dajt Mullaj <dajt.mullai@gmail.com>
 #
-# License: BSD 3 clause
+# License: BSD 3 clause 
 from __future__ import print_function
 
 import chemplot.descriptors as desc
@@ -16,12 +16,14 @@ import functools
 from sklearn.decomposition import PCA
 from sklearn.preprocessing import StandardScaler
 from sklearn.manifold import TSNE
+from sklearn.cluster import KMeans
+from pandas.api.types import is_numeric_dtype
 from rdkit.Chem import Draw
 from bokeh.plotting import figure
 from bokeh.transform import transform, factor_cmap
-from bokeh.palettes import Category10, Inferno
+from bokeh.palettes import Category10, Inferno, Spectral4
 from bokeh.models.mappers import LinearColorMapper
-from bokeh.models import ColorBar, HoverTool
+from bokeh.models import ColorBar, HoverTool, Panel, Tabs
 from bokeh.io import output_file, save, show
 from scipy import stats
 from io import BytesIO
@@ -59,6 +61,7 @@ class Plotter(object):
     :type __df_2_components: Dataframe
     :type __plot_title: string
     :type __data: list
+    :type pca_fit: sklearn.decomposition.TSNE
     :type tsne_fit: sklearn.manifold.TSNE
     :type umap_fit: umap.umap_.UMAP
     :type df_plot_xy: Dataframe
@@ -92,36 +95,52 @@ class Plotter(object):
         if self.__sim_type != "structural" and len(target) == 0:
             raise Exception("Target values missing")
         
-        # Error handeling target_type
+        # Error handeling target_type                
+        if len(target) > 0:
+            if len(target) != len(encoding_list):
+                raise Exception("If target is provided its length must match the instances of molecules")
+                
         if len(target) > 0:
             df_target = pd.DataFrame(data=target)
             unique_targets_ratio = 1.*df_target.iloc[:, 0].nunique()/df_target.iloc[:, 0].count() < 0.05
-            if target_type == 'R' and unique_targets_ratio:
+            numeric_target = is_numeric_dtype(df_target.dtypes[0])
+            if target_type == 'R' and (unique_targets_ratio or not numeric_target):
                 print('Input received is \'R\' for target values that seem not continuous.')
             if target_type not in self._target_types:
-                if unique_targets_ratio:
-                    self.__target_type = 'C'
-                    print('target_type indicates if the target is a continuous variable or a class label.\n'+
-                          'R stands for regression and C for classification. Input R as target type for continuous variables and C for class labels.\n'+
-                          'From analysis of the target, C has been selected for target_type.')
-                else:
+                if not unique_targets_ratio and numeric_target:
                     self.__target_type = 'R'
                     print('target_type indicates if the target is a continuous variable or a class label.\n'+
                           'R stands for regression and C for classification. Input R as target type for continuous variables and C for class labels.\n'+
                           'From analysis of the target, R has been selected for target_type.')
+                else:
+                    self.__target_type = 'C'
+                    print('target_type indicates if the target is a continuous variable or a class label.\n'+
+                          'R stands for regression and C for classification. Input R as target type for continuous variables and C for class labels.\n'+
+                          'From analysis of the target, C has been selected for target_type.')
             else:
                 self.__target_type = target_type
         else:
             self.__target_type = None
+        
+        if len(target) > 0 and self.__target_type == 'C':
+            df_target = pd.DataFrame(data=target)
+            if df_target.iloc[:, 0].nunique() == 1:
+                target = []
+                self.__sim_type = "structural"
+                print("Only one class found in the targets")
           
         # Instantiate Plotter class
         if self.__sim_type == "tailored":
             self.__mols, df_descriptors, target = get_desc(encoding_list, target)
+            if df_descriptors.empty:
+                raise Exception("Descriptors could not be computed for given molecules")
             self.__df_descriptors, self.__target = desc.select_descriptors_lasso(df_descriptors,target,kind=self.__target_type)
         elif self.__sim_type == "structural":
             self.__mols, self.__df_descriptors, self.__target = get_fingerprints(encoding_list,target,2,2048)
             
-            
+        if len(self.__mols) < 2 or len(self.__df_descriptors.columns) < 2:
+            raise Exception("Plotter object cannot be instantiated for given molecules")
+                
         self.__df_2_components = None
         self.__plot_title = None
             
@@ -166,17 +185,19 @@ class Plotter(object):
         return cls(inchi_list, target, target_type, sim_type, desc.get_mordred_descriptors_from_inchi, desc.get_ecfp_from_inchi)
         
     
-    def pca(self):
+    def pca(self, **kwargs):
         """
         Calculates the first 2 PCA components of the molecular descriptors.
         
+        :param kwargs: Other keyword arguments are passed down to sklearn.decomposition.PCA
+        :type kwargs: key, value mappings
         :returns: The dataframe containing the PCA components.
         :rtype: Dataframe
         """
         self.__data = self.__data_scaler()
         
         # Linear dimensionality reduction to 2 components by PCA
-        self.pca_fit = PCA(n_components=2)
+        self.pca_fit = PCA(n_components=2, **kwargs)
         first2ecpf_components = self.pca_fit.fit_transform(self.__data)
         coverage_components = self.pca_fit.explained_variance_ratio_
         
@@ -192,19 +213,21 @@ class Plotter(object):
         if len(self.__target) > 0: 
             self.__df_2_components['target'] = self.__target
         
-        return self.__df_2_components
+        return self.__df_2_components.copy()
     
     
-    def tsne(self, perplexity=None, pca=False, random_state=None):
+    def tsne(self, perplexity=None, pca=False, random_state=None, **kwargs):
         """
         Calculates the first 2 t-SNE components of the molecular descriptors.
         
         :param perplexity: perplexity value for the t-SNE model  
         :param pca: indicates if the features must be preprocessed by PCA
-        :param random_state: random seed that can be passed as a parameter for reproducing the same results        
+        :param random_state: random seed that can be passed as a parameter for reproducing the same results     
+        :param kwargs: Other keyword arguments are passed down to sklearn.manifold.TSNE
         :type perplexity: int
         :type pca: boolean
         :type random_state: int
+        :type kwargs: key, value mappings
         :returns: The dataframe containing the t-SNE components.
         :rtype: Dataframe
         """ 
@@ -233,7 +256,7 @@ class Plotter(object):
                 print('Robust results are obtained for values of perplexity between 5 and 50')
         
         # Embed the data in two dimensions
-        self.tsne_fit = TSNE(n_components=2, perplexity=perplexity, random_state=random_state)
+        self.tsne_fit = TSNE(n_components=2, perplexity=perplexity, random_state=random_state, **kwargs)
         ecfp_tsne_embedding = self.tsne_fit.fit_transform(self.__data)
         # Create a dataframe containinting the first 2 TSNE components of ECFP 
         self.__df_2_components = pd.DataFrame(data = ecfp_tsne_embedding
@@ -242,19 +265,21 @@ class Plotter(object):
         if len(self.__target) > 0: 
             self.__df_2_components['target'] = self.__target
         
-        return self.__df_2_components
+        return self.__df_2_components.copy()
         
         
-    def umap(self, n_neighbors=None, min_dist=None, pca=False, random_state=None):
+    def umap(self, n_neighbors=None, min_dist=None, pca=False, random_state=None, **kwargs):
         """
         Calculates the first 2 UMAP components of the molecular descriptors.
         
         :param num_neighbors: Number of neighbours used in the UMAP madel.
         :param min_dist: Value between 0.0 and 0.99, indicates how close to each other the points can be displayed.
         :param random_state: random seed that can be passed as a parameter for reproducing the same results
+        :param kwargs: Other keyword arguments are passed down to umap.UMAP
         :type num_neighbors: int
         :type min_dist: float
         :type random_state: int
+        :type kwargs: key, value mappings
         :returns: The dataframe containing the UMAP components.
         :rtype: Dataframe
         """  
@@ -289,7 +314,7 @@ class Plotter(object):
                 min_dist = parameters.MIN_DIST_TAILORED
             
         # Embed the data in two dimensions
-        self.umap_fit = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist, random_state=random_state, n_components=2)
+        self.umap_fit = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist, random_state=random_state, n_components=2, **kwargs)
         ecfp_umap_embedding = self.umap_fit.fit_transform(self.__data)
         # Create a dataframe containinting the first 2 UMAP components of ECFP 
         self.__df_2_components = pd.DataFrame(data = ecfp_umap_embedding
@@ -298,9 +323,36 @@ class Plotter(object):
         if len(self.__target) > 0: 
             self.__df_2_components['target'] = self.__target
         
-        return self.__df_2_components
+        return self.__df_2_components.copy()
     
-    def visualize_plot(self, size=20, kind="scatter", remove_outliers=False, is_colored=True, colorbar=False, filename=None, title=None):
+    
+    def cluster(self, n_clusters=5, **kwargs):
+        """
+        Computes the clusters presents in the embedded chemical space.
+        
+        :param n_clusters: Number of clusters that will be computed  
+        :param kwargs: Other keyword arguments are passed down to sklearn.cluster.KMeans
+        :type n_clusters: int
+        :type kwargs: key, value mappings
+        :returns: The dataframe containing the 2D embedding.
+        :rtype: Dataframe
+        """
+        if self.__df_2_components is None:
+            print('Reduce the dimensions of your molecules before clustering.')
+            return None
+        
+        x = self.__df_2_components.columns[0]
+        y = self.__df_2_components.columns[1]
+        
+        cluster = KMeans(n_clusters, **kwargs)
+        
+        cluster.fit(self.__df_2_components[[x,y]])
+        self.__df_2_components['clusters'] = cluster.labels_.tolist()
+        
+        return self.__df_2_components.copy()
+    
+    
+    def visualize_plot(self, size=20, kind="scatter", remove_outliers=False, is_colored=True, colorbar=False, clusters=False, filename=None, title=None):
         """
         Generates a plot for the given molecules embedded in two dimensions.
         
@@ -309,13 +361,17 @@ class Plotter(object):
         :param remove_outliers: Boolean value indicating if the outliers must be identified and removed 
         :param is_colored: Indicates if the points must be colored according to target 
         :param colorbar: Indicates if the plot legend must be represented as a colorbar. Only considered when the target_type is "R".
+        :param clusters: If True the clusters are shown instead of possible targets. Pass a list or a int to only show selected clusters (indexed by int).
         :param filename: Indicates the file where to save the plot
+        :param title: Title of the plot.
         :type size: int
         :type kind: string
         :type remove_outliers: boolean
         :type is_colored: boolean
         :type colorbar: boolean
+        :type clusters: boolean or list or int
         :type filename: string
+        :type title: string
         :returns: The matplotlib axes containing the plot.
         :rtype: Axes
         """
@@ -323,6 +379,9 @@ class Plotter(object):
             print('Reduce the dimensions of your molecules before creating a plot.')
             return None
         
+        if clusters is not False and 'clusters' not in self.__df_2_components:
+            print('Call cluster() before visualizing a plot with clusters.')
+        
         if title is None:
             title = self.__plot_title
         
@@ -339,15 +398,39 @@ class Plotter(object):
         
         # Define colors 
         hue = None
+        hue_order = None
         palette = None
-        if len(self.__target) == 0:
-            is_colored = False;
+        if clusters is not False and 'clusters' in self.__df_2_components.columns:
+            hue = 'clusters'
+            palette = 'deep'
+            if not isinstance(clusters, bool):
+                if isinstance(clusters, int): clusters = [clusters]
+                df_data['clusters'] = df_data['clusters'].isin(clusters)
+                # Labels cluster
+                total = df_data['clusters'].value_counts()
+                t_s = total.get(True) if total.get(True) else 0
+                p_s = t_s / total.sum()
+                p_o = 1 - p_s
+                labels = {
+                    True: f'Selected - {p_s:.0%}', 
+                    False: f'Other - {p_o:.0%}'
+                    }
+                df_data.clusters.replace(labels, inplace=True)
+                hue_order = list(labels.values())
+            else:
+                hue_order = self.__percentage_clusters(df_data)
+                hue_order.sort()
         else:
-            if is_colored:
-                df_data = df_data.assign(target=self.__target)
-                hue = 'target'
-                if self.__target_type == "R":
-                    palette = sns.color_palette("inferno", as_cmap=True)
+            if len(self.__target) == 0:
+                is_colored = False;
+            else:
+                if is_colored:
+                    df_data = df_data.assign(target=self.__target)
+                    hue = 'target'
+                    palette = 'deep'
+                    if self.__target_type == "R":
+                        palette = sns.color_palette("inferno", as_cmap=True)
+                        
         
         # Remove outliers (using Z-score)
         if remove_outliers:
@@ -360,9 +443,10 @@ class Plotter(object):
         
         # Create a plot based on the reduced components 
         if kind == "scatter":
-            plot = sns.scatterplot(x=x, y=y, hue=hue, palette=palette, data=df_data, s=size*3)
+            plot = sns.scatterplot(x=x, y=y, hue=hue, hue_order=hue_order, palette=palette, data=df_data, s=size*3)
             plot.set_label("scatter")
             axis = plot
+            plot.legend(markerscale=size*0.19)
             # Add colorbar
             if self.__target_type == "R" and colorbar:
                 plot.get_legend().remove()
@@ -396,7 +480,7 @@ class Plotter(object):
         
         return axis
     
-    def interactive_plot(self, size=700, kind="scatter", remove_outliers=False, is_colored=True, filename=None, show_plot=False, title=None,):
+    def interactive_plot(self, size=700, kind="scatter", remove_outliers=False, is_colored=True, clusters=False, filename=None, show_plot=False, title=None,):
         """
         Generates an interactive Bokeh plot for the given molecules embedded in two dimensions.
         
@@ -404,12 +488,18 @@ class Plotter(object):
         :param kind: Type of plot 
         :param remove_outliers: Boolean value indicating if the outliers must be identified and removed 
         :param is_colored: Indicates if the points must be colored according to target 
+        :param clusters: Indicates if to add a tab with the clusters if these have been computed
         :param filename: Indicates the file where to save the Bokeh plot
+        :param show_plot: Immediately display the current plot. 
+        :param title: Title of the plot.
         :type size: int
         :type kind: string
         :type remove_outliers: boolean
         :type is_colored: boolean
+        :type cluster: boolean
         :type filename: string
+        :type show_plot: boolean
+        :type title: string
         :returns: The bokeh figure containing the plot.
         :rtype: Figure
         """
@@ -417,6 +507,9 @@ class Plotter(object):
             print('Reduce the dimensions of your molecules before creating a plot.')
             return None
         
+        if clusters and 'clusters' not in self.__df_2_components:
+            print('Call cluster() before visualizing a plot with clusters.')
+            
         if title is None:
             title = self.__plot_title
         
@@ -441,9 +534,10 @@ class Plotter(object):
         # Remove outliers (using Z-score)
         if remove_outliers:
             df_data = self.__remove_outliers(x, y, df_data)
-            
+        
+        tabs = None
         if kind == "scatter":
-            p = self.__interactive_scatter(x, y, df_data, size, is_colored, title)
+            p, tabs = self.__interactive_scatter(x, y, df_data, size, is_colored, clusters, title)
         else: 
             p = self.__interactive_hex(x, y, df_data, size, title)
             
@@ -457,6 +551,9 @@ class Plotter(object):
         p.xaxis.major_label_text_font_size = '0pt'  
         p.yaxis.major_label_text_font_size = '0pt' 
         
+        if tabs is not None:
+            p = tabs
+            
         # Save plot
         if filename is not None:
             output_file(filename, title=title)
@@ -493,7 +590,25 @@ class Plotter(object):
         
         return df[filtered_entries]
     
-    def __interactive_scatter(self, x, y, df_data, size, is_colored, title):       
+    def __percentage_clusters(self, df_data):
+        total = df_data['clusters'].value_counts()
+        sum_tot = total.sum()
+        labels = {}
+        count = 0
+        for key, value in total.items():
+            p = float(f"{(value/sum_tot)*100:.0f}")
+            labels[key] = p
+            count += p
+        # Solve possible rounding errors
+        if 100 - count > 0:
+            labels[0] = labels[0] + 100 - count
+        for key, value in labels.items():
+            labels[key] = f"Cluster {key} - {value:.0f}%"
+        # Edit df_data and return labels
+        df_data.clusters.replace(labels, inplace=True)
+        return list(labels.values())
+            
+    def __interactive_scatter(self, x, y, df_data, size, is_colored, clusters, title):       
         # Add images column
         df_data['imgs'] = self.__mol_to_2Dimage(list(df_data['mols']))
         df_data.drop(columns=['mols'], inplace=True)
@@ -526,7 +641,36 @@ class Plotter(object):
                 color_bar = ColorBar(color_mapper=color_mapper, location=(0,0))
                 p.add_layout(color_bar, 'right')
         
-        return p
+        tabs = None
+        if clusters and 'clusters' in df_data.columns:
+            p_c = figure(title=title, plot_width=size, plot_height=size, tools=tools, tooltips=parameters.TOOLTIPS_CLUSTER)
+            # Get percentages
+            self.__percentage_clusters(df_data)
+            clusters = df_data.groupby(['clusters'])
+            for cluster, color in zip(clusters, Category10[10]):
+                p_c.circle(x=x, y=y, size=2.5, alpha=1, line_color=color, fill_color=color,
+                     legend_label=f'{cluster[0]}', muted_color=('#717375'), muted_alpha=0.2,
+                     source=cluster[1])
+                
+            p_c.legend.location = "top_left"
+            p_c.legend.title = "Clusters"
+            p_c.legend.click_policy = "mute"
+            
+            p_c.xaxis[0].axis_label = x
+            p_c.yaxis[0].axis_label = y
+            
+            p_c.xaxis.major_tick_line_color = None  
+            p_c.xaxis.minor_tick_line_color = None
+            p_c.yaxis.major_tick_line_color = None  
+            p_c.yaxis.minor_tick_line_color = None 
+            p_c.xaxis.major_label_text_font_size = '0pt'  
+            p_c.yaxis.major_label_text_font_size = '0pt' 
+            
+            tab1 = Panel(child=p, title="Plot")
+            tab2 = Panel(child=p_c, title="Clusters")
+            tabs = Tabs(tabs=[tab1, tab2])
+        
+        return p, tabs
                 
     def __interactive_hex(self, x, y, df_data, size, title):      
         # Hex Plot
